/********************************************************/
/* Virtual Applications' OpenGL Renderer Source Code    */
/********************************************************/
/* You're allowed to make changes to this file. If the  */
/* changes are constructive, you can e-mail me with the */
/* patch to be applied to this source. You'll be given  */
/* full credits for changes you've made.                */
/********************************************************/
/* Contributors:                                        */
/* Jean-Philip Desjardins (First Code and Maintainer)   */
/* Yves Rioux (Some Sprite Fixes and Fl/Cl Code)		*/
/********************************************************/

#include "COMMON.H"

struct VOGL_WALLTEXTURE
{
	float	U1;
	float	U2;
	float	V1_Main;
	float	V2_Main;
	float	V1_Top;
	float	V2_Top;
	float	V1_Bottom;
	float	V2_Bottom;
};

struct VOGL_SPRITETEXTURE
{
	float U_Top;
	float V_Top;
	float U_Bottom;
	float V_Bottom;
};

extern HDC				VOGL_hDC;
XBUILD_MAP				VOGL_Map;
XBUILD_ART_TILEDATA*	VOGL_Tiles[MAXTILES];
int						VOGL_nCurrentParallax = 0;
int						VOGL_TilesCreated[MAXTILES];
unsigned _int8*			VOGL_lpPalette = NULL;
unsigned int			VOGL_nTexture[MAXTILES];
VOGL_WALLTEXTURE		VOGL_WallTexture[MAXWALLS];
VOGL_SPRITETEXTURE		VOGL_SpriteTexture[MAXSPRITES];
extern VOGL_CONFIG		VOGL_Config;
extern int				VOGL_nMesh[MAXTILES];
int						VOGL_SpriteMax;
int						VOGL_SpriteMin;
LISTSPRITE*				VOGL_pSpriteBuffer = NULL;
extern int				nTriangles;
extern int				VOGL_Width[];
extern int				VOGL_Height[];
extern int				VOGL_nSectorDL[MAXSECTORS][2];

void CreateDisplayListForSector(int);

int VOGL_nDistances[3] =
{
	60000,
	20000,
	10000
};

#define PI 3.141592

//Should normalize that :P
__inline int SearchForNextPower(int nNumber)
{
	int power;
	for(power = 2; power < 1024; power *= 2)
	{
		if(power >= nNumber)
		{
			return power;
		}
	}
	return 0;
}

/*__inline int SearchForNextPower(int nNumber)
{
	int power;
	if(nNumber >= 256)
	{
		//Err... 3DFX hack
		return 256;
	}
	for(power = 2; power < 1024; power *= 2)
	{
		if(power >= nNumber)
		{
			return power;
		}
	}
	return 0;
}*/

void PreCalculate()
{
	int i, a, k;
	memset(VOGL_WallTexture, 0, sizeof(VOGL_WALLTEXTURE) * 8192);
	for(i = 0; i < VOGL_Map.Mapinfo.WallNumber; i++)
	{
		float wallheight;
		int j;
		for(k = 0; k < VOGL_Map.Mapinfo.SectorNumber; k++)
		{
			for(j = 0; j < VOGL_Map.Sector[k].WallNumber; j++)
			{
				if(i == VOGL_Map.Sector[k].WallPointer + j)
				{
					a = k;
					break;
				}
			}
		}
		VOGL_WallTexture[i].U1 = (float)VOGL_Map.Wall[i].PanX / (float)VOGL_Tiles[VOGL_Map.Wall[i].Pic]->Width;
		VOGL_WallTexture[i].U2 = (((float)VOGL_Map.Wall[i].RepX * 8.0f) / (float)VOGL_Tiles[VOGL_Map.Wall[i].Pic]->Width) + VOGL_WallTexture[i].U1;
		
		if(VOGL_Map.Wall[i].CStat & 0x8)
		{
			wallheight = VOGL_WallTexture[i].U2;
			VOGL_WallTexture[i].U2 = VOGL_WallTexture[i].U1;
			VOGL_WallTexture[i].U1 = wallheight;
		}
		wallheight = (float)VOGL_Map.Sector[a].CZ - (float)VOGL_Map.Sector[a].FZ;
		wallheight /= 16;
		VOGL_WallTexture[i].V1_Main = (float)(VOGL_Map.Wall[i].PanY % VOGL_Tiles[VOGL_Map.Wall[i].Pic]->Height);
		VOGL_WallTexture[i].V2_Main = (((float)wallheight / (float)VOGL_Tiles[VOGL_Map.Wall[i].Pic]->Height) * ((float)VOGL_Map.Wall[i].RepY / 8.0f)) + VOGL_WallTexture[i].V1_Main;
		if(VOGL_Map.Wall[i].NextSector != -1 || VOGL_Map.Wall[i].NextWall != 0xFFFF)
		{
			wallheight = (float)VOGL_Map.Sector[a].CZ - (float)VOGL_Map.Sector[VOGL_Map.Wall[i].NextSector].CZ; 
			wallheight /= 16.0f;
			VOGL_WallTexture[i].V1_Top = (float)(VOGL_Map.Wall[i].PanY % VOGL_Tiles[VOGL_Map.Wall[i].Pic]->Height);
			VOGL_WallTexture[i].V2_Top = ((float)wallheight / (float)VOGL_Tiles[VOGL_Map.Wall[i].Pic]->Height * (float)VOGL_Map.Wall[i].RepY / 8.0f) + VOGL_WallTexture[i].V1_Top;
			if(VOGL_Map.Wall[i].CStat & 0x100)
			{
				wallheight = VOGL_WallTexture[i].V2_Top;
				VOGL_WallTexture[i].V2_Top = VOGL_WallTexture[i].V1_Top;
				VOGL_WallTexture[i].V1_Top = wallheight;
			}
			wallheight = (float)VOGL_Map.Sector[VOGL_Map.Wall[i].NextSector].FZ - (float)VOGL_Map.Sector[a].FZ;
			wallheight /= 16.0f;
			VOGL_WallTexture[i].V1_Bottom = (float)(VOGL_Map.Wall[i].PanY % VOGL_Tiles[VOGL_Map.Wall[i].Pic]->Height);
			VOGL_WallTexture[i].V2_Bottom = ((float)wallheight / (float)VOGL_Tiles[VOGL_Map.Wall[i].Pic]->Height * (float)VOGL_Map.Wall[i].RepY / 8.0f) + VOGL_WallTexture[i].V1_Bottom;
			if(VOGL_Map.Wall[i].CStat & 0x100)
			{
				wallheight = VOGL_WallTexture[i].V2_Bottom;
				VOGL_WallTexture[i].V2_Bottom = VOGL_WallTexture[i].V1_Bottom;
				VOGL_WallTexture[i].V1_Bottom = wallheight;
			}
		}
	}
}

void UploadTexture(int nTile)
{
	int nWidth;
	int nHeight;
	unsigned _int8 *scaled;
	int i, j;
	int artptr = 0;
	unsigned _int8 *lpData;
	if(VOGL_lpPalette == NULL)
	{
		VOGL_lpPalette = new unsigned _int8[0x300];
	}
	XBUILD_ART_GetPalette(VOGL_lpPalette);
	nWidth = VOGL_Tiles[nTile]->Width;
	nHeight = VOGL_Tiles[nTile]->Height;
	lpData = new unsigned _int8[VOGL_Tiles[nTile]->Width * VOGL_Tiles[nTile]->Height * 4];
	for(i = 0; i < VOGL_Tiles[nTile]->Width; i++)
	{
		for(j = VOGL_Tiles[nTile]->Height; j > 0; j--)
		{
			if(VOGL_Tiles[nTile]->lpData[artptr] == 0xFF)
			{
				lpData[((((j - 1) * nWidth) + i) * 4) + 0] = 0;
				lpData[((((j - 1) * nWidth) + i) * 4) + 1] = 0;
				lpData[((((j - 1) * nWidth) + i) * 4) + 2] = 0;
				lpData[((((j - 1) * nWidth) + i) * 4) + 3] = 0;
			}
			else
			{
				lpData[((((j - 1) * nWidth) + i) * 4) + 0] = (VOGL_lpPalette[((VOGL_Tiles[nTile]->lpData[artptr]) * 3) + 0]) << 2;
				lpData[((((j - 1) * nWidth) + i) * 4) + 1] = (VOGL_lpPalette[((VOGL_Tiles[nTile]->lpData[artptr]) * 3) + 1]) << 2;
				lpData[((((j - 1) * nWidth) + i) * 4) + 2] = (VOGL_lpPalette[((VOGL_Tiles[nTile]->lpData[artptr]) * 3) + 2]) << 2;
				lpData[((((j - 1) * nWidth) + i) * 4) + 3] = 0xFF;
			}
			artptr++;
		}
	}
	glGenTextures(1, &VOGL_nTexture[nTile]);
	glBindTexture(GL_TEXTURE_2D, VOGL_nTexture[nTile]);
	if(!VOGL_Config.nMipmapping)
	{
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	}
	else
	{
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	}
	scaled = new unsigned _int8[SearchForNextPower(VOGL_Tiles[nTile]->Width) * SearchForNextPower(VOGL_Tiles[nTile]->Height) * 4];
	gluScaleImage(GL_RGBA, VOGL_Tiles[nTile]->Width, VOGL_Tiles[nTile]->Height, GL_UNSIGNED_BYTE, lpData, SearchForNextPower(VOGL_Tiles[nTile]->Width), SearchForNextPower(VOGL_Tiles[nTile]->Height), GL_UNSIGNED_BYTE, scaled); 
	if(!VOGL_Config.nMipmapping)
	{
		glTexImage2D(GL_TEXTURE_2D, 0, 4, SearchForNextPower(VOGL_Tiles[nTile]->Width), SearchForNextPower(VOGL_Tiles[nTile]->Height), 0, GL_RGBA, GL_UNSIGNED_BYTE, scaled);
	}
	else
	{
		gluBuild2DMipmaps(GL_TEXTURE_2D, 4, SearchForNextPower(VOGL_Tiles[nTile]->Width), SearchForNextPower(VOGL_Tiles[nTile]->Height), GL_RGBA, GL_UNSIGNED_BYTE, scaled);
	}
	delete scaled;
	delete lpData;
}

extern "C" void PrepareMap(void *lpData)
{
	if(VOGL_lpPalette != NULL)
	{
		delete VOGL_lpPalette;
	}
	int i;
	int nTile;
	memmove(&VOGL_Map, lpData, sizeof(XBUILD_MAP));	
	VOGL_lpPalette = new unsigned _int8[0x300];
	for(i = 0; i < VOGL_Map.Mapinfo.WallNumber; i++)
	{
		nTile = VOGL_Map.Wall[i].Pic;
		if(!VOGL_TilesCreated[nTile])
		{
			VOGL_Tiles[nTile] = new XBUILD_ART_TILEDATA;
			XBUILD_ART_GetTileData(nTile, VOGL_Tiles[nTile]);
			UploadTexture(nTile);
			VOGL_TilesCreated[nTile] = 0xFF;
		}
	}
	for(i = 0; i < VOGL_Map.Mapinfo.SectorNumber; i++)
	{
		nTile = VOGL_Map.Sector[i].CPic;
		if(!VOGL_TilesCreated[nTile])
		{
			VOGL_Tiles[nTile] = new XBUILD_ART_TILEDATA;
			XBUILD_ART_GetTileData(nTile, VOGL_Tiles[nTile]);
			UploadTexture(nTile);
			VOGL_TilesCreated[nTile] = 0xFF;
		}
		nTile = VOGL_Map.Sector[i].FPic;
		if(!VOGL_TilesCreated[nTile])
		{
			VOGL_Tiles[nTile] = new XBUILD_ART_TILEDATA;
			XBUILD_ART_GetTileData(nTile, VOGL_Tiles[nTile]);
			UploadTexture(nTile);
			VOGL_TilesCreated[nTile] = 0xFF;
		}
		CreateDisplayListForSector(i);
	}
	for(i = 0; i < VOGL_Map.Mapinfo.SpriteNumber; i++)
	{
		nTile = VOGL_Map.Sprite[i].Pic;
		if(!VOGL_TilesCreated[nTile])
		{
			VOGL_Tiles[nTile] = new XBUILD_ART_TILEDATA;
			XBUILD_ART_GetTileData(nTile, VOGL_Tiles[nTile]);
			UploadTexture(nTile);
			VOGL_TilesCreated[nTile] = 0xFF;
		}
	}
	VOGL_pSpriteBuffer = new LISTSPRITE[60001];
	PreCalculate();
}

__inline int CanDrawSector(int nSector, int x, int y)
{
	if(VOGL_Config.nDistance == 0)
	{
		return 1;
	}
	int Debut = VOGL_Map.Sector[nSector].WallPointer;
	for(int i=0; i < VOGL_Map.Sector[nSector].WallNumber ; i++)
	{
		int	WX = VOGL_Map.Wall[Debut++].X;
		int	WY = VOGL_Map.Wall[Debut++].Y;

		int resultX = WX - x;
		int resultY = WY - y;
		
		int dist = (int)sqrt( (resultX*resultX) + (resultY*resultY) );

		if( dist < VOGL_nDistances[VOGL_Config.nDistance] )
			return 1;
	}
	return 0;
}

__inline int CanDrawSprite(int nSprite, int x, int y)
{
	if(VOGL_Config.nDistance == 0)
	{
		return 1;
	}
	int	WX = VOGL_Map.Sprite[nSprite].X;
	int	WY = VOGL_Map.Sprite[nSprite].Y;
	
	int resultX = WX - x;
	int resultY = WY - y;
		
	int dist = (int)sqrt( (resultX*resultX) + (resultY*resultY) );

	if( dist < VOGL_nDistances[VOGL_Config.nDistance] )
		return 1;

	return 0;
}

__inline void SortSprites(int x, int y,int* max, int* min, LISTSPRITE *pBuffer)
{
	int k, j;
	int nCount = 0;
	int Max = 0;
	int Min = VOGL_nDistances[VOGL_Config.nDistance];
	static LISTSPRITE pPrimary[60001];
	memset(pBuffer, 0, sizeof(LISTSPRITE) * 60001);
	memset(pPrimary, 0, sizeof(LISTSPRITE) * 60001);
	for(k = 0; k < VOGL_Map.Mapinfo.SpriteNumber; k++)
	{
		if((VOGL_Map.Sprite[k].CStat & 0x30) == 0x00 && CanDrawSprite(k,x,y))
		{
			int SommeX,SommeY;
			SommeX = x-VOGL_Map.Sprite[k].X;
			SommeY = y-VOGL_Map.Sprite[k].Y;
			pPrimary[nCount].Distance = (int)sqrt( SommeX*SommeX + SommeY*SommeY );
			pPrimary[nCount].Donnee = (int*)&VOGL_Map.Sprite[k];
			pPrimary[nCount].nSprite = k;
			if( Max < pPrimary[nCount].Distance)
			{
				Max = pPrimary[nCount].Distance;
			}
			if( Min > pPrimary[nCount].Distance)
			{
				Min = pPrimary[nCount].Distance;
			}
			nCount++;
		}
	}
	*min = Min;
	*max = Max;
	for(j = 0; j < nCount; j++)
	{
		if(pBuffer[pPrimary[j].Distance].Donnee == NULL)
		{
			pBuffer[pPrimary[j].Distance].Donnee = pPrimary[j].Donnee;
			pBuffer[pPrimary[j].Distance].nSprite = pPrimary[j].nSprite;
		}
		else
		{
			if(pBuffer[pPrimary[j].Distance+1].Donnee == NULL)
			{
				pBuffer[pPrimary[j].Distance+1].Donnee = pPrimary[j].Donnee;
				pBuffer[pPrimary[j].Distance+2].nSprite = pPrimary[j].nSprite;
			}
			else
			{
				pBuffer[pPrimary[j].Distance+2].Donnee = pPrimary[j].Donnee;
				pBuffer[pPrimary[j].Distance+2].nSprite = pPrimary[j].nSprite;
			}
		}
	}
}

void DrawParallax(int nYRot, int nXRot)
{	
	float fYRatio, fXRatio;
	if(VOGL_nCurrentParallax == 0)
	{
		return;
	}
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glOrtho(0, 320, 0, 200, -1, 1);
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	glColor3f(1.0f, 1.0f, 1.0f);
	glTranslated(0, 0, 1);
	fYRatio = (float)nYRot;
	fYRatio /= 4096;
	fXRatio = (float)nXRot;
	fXRatio /= 2048;
	if(VOGL_nCurrentParallax)
	{
		glBindTexture(GL_TEXTURE_2D, VOGL_nTexture[VOGL_nCurrentParallax]);
	}
	glDisable(GL_DEPTH_TEST);
	glBegin(GL_QUADS);
	{
		glTexCoord2f(fYRatio,			fXRatio);			glVertex2i(0,		40);
		glTexCoord2f(3.0f + fYRatio,	fXRatio);			glVertex2i(0 + 320,	40);
		glTexCoord2f(3.0f + fYRatio,	1.0f + fXRatio);	glVertex2i(0 + 320,	40 + 200);
		glTexCoord2f(fYRatio,			1.0f + fXRatio);	glVertex2i(0,		40 + 200);
	}
	glEnd();
	glEnable(GL_DEPTH_TEST);
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
}

__inline void SetColor(char nShade, char nPalette)
{
	float shade, r, b, g;
	if(nShade >= 0x1C)
	{
		shade = 0.0f;
	}
	else
	{
		shade = 28.0f - (float)nShade;
		shade /= 28.0f;
	}
	switch(nPalette)
	{
	case 2:
		r = 1.0f; g = 0.0f; b = 0.0f;
		break;
	case 21:
		r = 1.0f; g = 0.0f; b = 0.0f;
		break;
	case 23:
		r = 1.0f; g = 1.0f; b = 0.0f;
		break;
	default:
		r = 1.0f; g = 1.0f; b = 1.0f;
		break;
	}
	glColor3f(shade * r, shade * g, shade * b);
}

__inline void DrawSprite(int k, int nx, int ny, int nz, int yrot)
{
	int x;
	int y;
	int z;
	int sizex;
	int sizey;
	int nNumber;
	if(!(VOGL_Map.Sprite[k].CStat & 0x8000) && (VOGL_Map.Sprite[k].Pic != 0xFFFF))
	{
		x =  (VOGL_Map.Sprite[k].X - nx);
		y =  (VOGL_Map.Sprite[k].Y - nz);
		z =  -(VOGL_Map.Sprite[k].Z - ny);
		if(VOGL_nMesh[VOGL_Map.Sprite[k].Pic] != -1)
		{
			glPushMatrix();
			glTranslatef(x, z, y);
			glRotatef(-90.0f, 1.0f, 0.0f, 0.0f);
			glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
			glBindTexture(GL_TEXTURE_2D, 0);
			glCallList(VOGL_nMesh[VOGL_Map.Sprite[k].Pic]);
			glPopMatrix();
			return;
		}
		SetColor(VOGL_Map.Sprite[k].Shade, VOGL_Map.Sprite[k].Palette);
		nNumber = VOGL_Map.Sprite[k].Pic;
		if(!VOGL_TilesCreated[nNumber])
		{
			VOGL_Tiles[nNumber] = new XBUILD_ART_TILEDATA;
			XBUILD_ART_GetTileData(nNumber, VOGL_Tiles[nNumber]);
			UploadTexture(nNumber);
			VOGL_TilesCreated[nNumber] = 0xFF;
		}
		if(VOGL_Tiles[nNumber]->Attributes & 0xC0)
		{
			nNumber += XBUILD_ANIMATE_GetTileIncrement(VOGL_Tiles[nNumber]);
		}
		if(!VOGL_TilesCreated[nNumber])
		{
			VOGL_Tiles[nNumber] = new XBUILD_ART_TILEDATA;
			XBUILD_ART_GetTileData(nNumber, VOGL_Tiles[nNumber]);
			UploadTexture(nNumber);
			VOGL_TilesCreated[nNumber] = 0xFF;
		}
		sizex = (VOGL_Tiles[nNumber]->Width);
		sizey = (VOGL_Tiles[nNumber]->Height);
		glPushMatrix();
		if((VOGL_Map.Sprite[k].CStat & 0x30) == 0x10)
		{
			glTranslated(x, z, y);
			glRotated(XBUILD_ANIMATE_BuildToDegree(VOGL_Map.Sprite[k].Angle), 0.0f, 1.0f, 0.0f);
			glTranslated(-x, -z, -y);
		}
		else if((VOGL_Map.Sprite[k].CStat & 0x30) == 0x20)
		{
			glTranslated(x, z, y);
			glRotated(XBUILD_ANIMATE_BuildToDegree(VOGL_Map.Sprite[k].Angle), 0.0f, 1.0f, 0.0f);
			glRotated(90, 1.0f, 0.0f, 0.0f);
			glTranslated(-x, -z, -y);
		}
		else
		{
			glTranslated(x, z, y);
			glRotated(XBUILD_ANIMATE_BuildToDegree(yrot) + 180.0f, 0.0f, 1.0f, 0.0f);
			glTranslated(-x, -z, -y);
		}
		glBindTexture(GL_TEXTURE_2D, VOGL_nTexture[nNumber]);
		glBegin(GL_QUADS);
		{
			glTexCoord2f(0.0f, 0.0f);	glVertex3i(x +   (sizex * VOGL_Map.Sprite[k].RepX / 8), z, y);
			glTexCoord2f(1.0f, 0.0f);	glVertex3i(x +  -(sizex * VOGL_Map.Sprite[k].RepX / 8), z, y);
			glTexCoord2f(1.0f, 1.0f);	glVertex3i(x +  -(sizex * VOGL_Map.Sprite[k].RepX / 8), z +  (sizey * VOGL_Map.Sprite[k].RepY / 4), y);
			glTexCoord2f(0.0f, 1.0f);	glVertex3i(x +   (sizex * VOGL_Map.Sprite[k].RepX / 8),	z +  (sizey * VOGL_Map.Sprite[k].RepY / 4), y);
		}
		glEnd();
		glPopMatrix();
	}
}

__inline void DrawSectorList(int sectnum, int nx, int ny, int nz, int nFloor)
{
	glPushMatrix();
	glTranslatef((float)-nx, (float)ny, (float)-nz);
	if(nFloor)
	{
		glCallList(VOGL_nSectorDL[sectnum][0]);
	}
	else
	{
		glCallList(VOGL_nSectorDL[sectnum][1]);
	}
	glPopMatrix();
}

extern "C" void DrawMap(int nx, int ny, int nz, int xrot, int yrot)
{
	int a, k, m, l, n;
	int x, x2, y, y2, z, z2, z3, z4;
	int nNumber;
	DrawParallax(yrot, xrot);
	glLoadIdentity();
	glRotated(xrot, 1, 0, 0);
	glRotated(-XBUILD_ANIMATE_BuildToDegree(yrot), 0, 1, 0);
	glColor3f(1.0f, 1.0f, 1.0f);
	glScalef(0.0625f, 0.0625f, 0.07f);
	glBindTexture(GL_TEXTURE_2D, NULL);
	for(a = 0; a < VOGL_Map.Mapinfo.SectorNumber; a++)
	{
		if( CanDrawSector(a,nx,nz) )
		{
			if(!(VOGL_Map.Sector[a].CStat & 1))
			{
				k = VOGL_Map.Sector[a].WallPointer;
				m = k;
				SetColor(VOGL_Map.Sector[a].CShade, VOGL_Map.Sector[k].CPal);
				nNumber = VOGL_Map.Sector[a].CPic;
				if(VOGL_Tiles[nNumber]->Attributes & 0xC0)
				{
					nNumber += XBUILD_ANIMATE_GetTileIncrement(VOGL_Tiles[nNumber]);
				}
				if(!VOGL_TilesCreated[nNumber])
				{
					VOGL_Tiles[nNumber] = new XBUILD_ART_TILEDATA;
					XBUILD_ART_GetTileData(nNumber, VOGL_Tiles[nNumber]);
					UploadTexture(nNumber);
					VOGL_TilesCreated[nNumber] = 0xFF;
				}
				glBindTexture(GL_TEXTURE_2D, VOGL_nTexture[nNumber]);
				DrawSectorList(a, nx, ny, nz, FALSE);
			}
			else
			{
				if(XBUILD_ANIMATE_GetCurrentSector() == a)
				{
					VOGL_nCurrentParallax = VOGL_Map.Sector[a].CPic;
				}
			}
			if(!(VOGL_Map.Sector[a].FStat & 1))
			{
				k = VOGL_Map.Sector[a].WallPointer;
				m = k;
				SetColor(VOGL_Map.Sector[a].FShade, VOGL_Map.Sector[k].FPal);
				nNumber = VOGL_Map.Sector[a].FPic;
				if(VOGL_Tiles[nNumber]->Attributes & 0xC0)
				{
					nNumber += XBUILD_ANIMATE_GetTileIncrement(VOGL_Tiles[nNumber]);
				}
				if(!VOGL_TilesCreated[nNumber])
				{
					VOGL_Tiles[nNumber] = new XBUILD_ART_TILEDATA;
					XBUILD_ART_GetTileData(nNumber, VOGL_Tiles[nNumber]);
					UploadTexture(nNumber);
					VOGL_TilesCreated[nNumber] = 0xFF;
				}
				glBindTexture(GL_TEXTURE_2D, VOGL_nTexture[nNumber]);
				DrawSectorList(a, nx, ny, nz, TRUE);
			}
			if(!(VOGL_Map.Sector[a].FStat & 1) || !(VOGL_Map.Sector[a].CStat & 1))
			{
				k = VOGL_Map.Sector[a].WallPointer;
				m = k;
				for(l = 0; l < VOGL_Map.Sector[a].WallNumber; l++)
				{
					SetColor(VOGL_Map.Wall[k].Shade, VOGL_Map.Wall[k].Pal);
					nNumber = VOGL_Map.Wall[k].Pic;
					if(VOGL_Tiles[nNumber]->Attributes & 0xC0)
					{
						nNumber += XBUILD_ANIMATE_GetTileIncrement(VOGL_Tiles[nNumber]);
					}
					if(!VOGL_TilesCreated[nNumber])
					{
						VOGL_Tiles[nNumber] = new XBUILD_ART_TILEDATA;
						XBUILD_ART_GetTileData(nNumber, VOGL_Tiles[nNumber]);
						UploadTexture(nNumber);
						VOGL_TilesCreated[nNumber] = 0xFF;
					}
					glBindTexture(GL_TEXTURE_2D, VOGL_nTexture[nNumber]);
					x  =  (VOGL_Map.Wall[k].X - nx);
					y  =  (VOGL_Map.Wall[k].Y - nz);
					x2 =  (VOGL_Map.Wall[VOGL_Map.Wall[k].Point2].X - nx);
					y2 =  (VOGL_Map.Wall[VOGL_Map.Wall[k].Point2].Y - nz);
					z  =  -(XBUILD_ANIMATE_GetCeilingHeight(a, VOGL_Map.Wall[k].X, VOGL_Map.Wall[k].Y) - ny);
					z2 =  -(XBUILD_ANIMATE_GetFloorHeight(a, VOGL_Map.Wall[k].X, VOGL_Map.Wall[k].Y) - ny);
					z3 =  -(XBUILD_ANIMATE_GetCeilingHeight(a, VOGL_Map.Wall[VOGL_Map.Wall[k].Point2].X, VOGL_Map.Wall[VOGL_Map.Wall[k].Point2].Y) - ny);
					z4 =  -(XBUILD_ANIMATE_GetFloorHeight(a, VOGL_Map.Wall[VOGL_Map.Wall[k].Point2].X, VOGL_Map.Wall[VOGL_Map.Wall[k].Point2].Y) - ny);
					//z  = -(VOGL_Map.Sector[a].CZ - ny);
					//z2 = -(VOGL_Map.Sector[a].FZ - ny);
					if(VOGL_Map.Wall[k].NextWall == 0xFFFF && VOGL_Map.Wall[k].NextSector == -1)
					{
						glBegin(GL_QUADS);
						{
							//glTexCoord2f(VOGL_WallTexture[k].U1,	VOGL_WallTexture[k].V1_Main);	glVertex3i(x , z , y );
							//glTexCoord2f(VOGL_WallTexture[k].U2,	VOGL_WallTexture[k].V1_Main);	glVertex3i(x2, z , y2);
							//glTexCoord2f(VOGL_WallTexture[k].U2,	VOGL_WallTexture[k].V2_Main);	glVertex3i(x2, z2, y2);
							//glTexCoord2f(VOGL_WallTexture[k].U1,	VOGL_WallTexture[k].V2_Main);	glVertex3i(x , z2, y );
							glTexCoord2f(VOGL_WallTexture[k].U1,	VOGL_WallTexture[k].V1_Main);	glVertex3i(x , z , y );
							glTexCoord2f(VOGL_WallTexture[k].U2,	VOGL_WallTexture[k].V1_Main);	glVertex3i(x2, z3, y2);
							glTexCoord2f(VOGL_WallTexture[k].U2,	VOGL_WallTexture[k].V2_Main);	glVertex3i(x2, z4, y2);
							glTexCoord2f(VOGL_WallTexture[k].U1,	VOGL_WallTexture[k].V2_Main);	glVertex3i(x , z2, y );
						}
						glEnd();
					}
					else
					{
						z  = -(VOGL_Map.Sector[a].CZ - ny);
						z2 = -(VOGL_Map.Sector[a].FZ - ny);
						z3 = -(VOGL_Map.Sector[VOGL_Map.Wall[k].NextSector].CZ - ny);
						z4 = -(VOGL_Map.Sector[VOGL_Map.Wall[k].NextSector].FZ - ny);
						if(z4 != z2)
						{
							if(z4 > z2)
							{
								glBegin(GL_QUADS);
								{
									glTexCoord2f(VOGL_WallTexture[k].U1, VOGL_WallTexture[k].V2_Bottom);	glVertex3i(x , z2, y );
									glTexCoord2f(VOGL_WallTexture[k].U2, VOGL_WallTexture[k].V2_Bottom);	glVertex3i(x2, z2, y2);
									glTexCoord2f(VOGL_WallTexture[k].U2, VOGL_WallTexture[k].V1_Bottom);	glVertex3i(x2, z4, y2);
									glTexCoord2f(VOGL_WallTexture[k].U1, VOGL_WallTexture[k].V1_Bottom);	glVertex3i(x , z4, y );
								}
								glEnd();
							}
						}
						if(z3 != z)
						{
							if(z3 < z)
							{
								glBegin(GL_QUADS);
								{
									glTexCoord2f(VOGL_WallTexture[k].U1, VOGL_WallTexture[k].V1_Top);	glVertex3i(x , z , y );
									glTexCoord2f(VOGL_WallTexture[k].U2, VOGL_WallTexture[k].V1_Top);	glVertex3i(x2, z , y2);
									glTexCoord2f(VOGL_WallTexture[k].U2, VOGL_WallTexture[k].V2_Top);	glVertex3i(x2, z3, y2);
									glTexCoord2f(VOGL_WallTexture[k].U1, VOGL_WallTexture[k].V2_Top);	glVertex3i(x , z3, y );
								}
								glEnd();
							}
						}	
					}
					n = k;
					k = VOGL_Map.Wall[k].Point2;
					if(k == m)
					{
						k = n + 1;
						m = k;
					}
				}
			}
		}	
	}
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);	
	for(k = 0; k < VOGL_Map.Mapinfo.SpriteNumber; k++)
	{
		if(CanDrawSprite(k,nx,nz))
		{
			if(VOGL_Map.Sprite[k].CStat & 0x30)
			{
				DrawSprite(k, nx, ny, nz, yrot);
			}
		}
	}	
	SortSprites(nx, nz, &VOGL_SpriteMax, &VOGL_SpriteMin, VOGL_pSpriteBuffer);
	for(k = VOGL_SpriteMax; k >= VOGL_SpriteMin; k--)
	{		
		if(VOGL_pSpriteBuffer[k].Donnee != NULL)
		{
			DrawSprite(VOGL_pSpriteBuffer[k].nSprite, nx, ny, nz, yrot);
		}
	}
	glDisable(GL_BLEND);
}

extern "C" void SwapDisplayBuffers()
{
	SwapBuffers(VOGL_hDC);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glLoadIdentity();
}

extern "C" void ProcessDisplayList(void *d)
{
	XBUILDDISPLAYLIST *l = (XBUILDDISPLAYLIST*)d;
	int i;
	int nTile;
	int nHeight;
	VOGL_lpPalette = new unsigned _int8[0x300];
	glDisable(GL_DEPTH_TEST);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	//Otrho Screen Init
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	if(l->dwReferentialX != 0xFFFFFFFF)
	{
		glOrtho(0, l->dwReferentialX, 0, l->dwReferentialY, -1, 1);
		nHeight = l->dwReferentialY;
	}
	else
	{
		glOrtho(0, VOGL_Width[VOGL_Config.nResolution], 0, VOGL_Height[VOGL_Config.nResolution], -1, 1);
		nHeight = VOGL_Height[VOGL_Config.nResolution];
	}
	glMatrixMode(GL_MODELVIEW);            
	glPushMatrix();
	//Ortho Screen Init
	glLoadIdentity();
	glColor3f(1.0f, 1.0f, 1.0f);
	for(i = 0; i < l->nElementCount; i++)
	{
		switch(l->Element[i].dwType)
		{
		case XBUILDDISPLAYLIST_TYPE_FILTER:
			float fR, fG, fB, fA;
			XBUILDDISPLAYELEMENT_FILTER Filter;
			memcpy(&Filter, l->Element[i].pData, sizeof(XBUILDDISPLAYELEMENT_FILTER));
			glBindTexture(GL_TEXTURE_2D, NULL);
			fR = (float)Filter.R / 255.0f;
			fG = (float)Filter.G / 255.0f;
			fB = (float)Filter.B / 255.0f;
			fA = (float)Filter.A / 255.0f;
			glColor4f(fR, fG, fB, fA);
			glBegin(GL_QUADS);
			{
				glVertex2i(0,					l->dwReferentialY);
				glVertex2i(l->dwReferentialX,	l->dwReferentialY);
				glVertex2i(l->dwReferentialX,	0);
				glVertex2i(0,					0);
			}
			glEnd();
			break;
		case XBUILDDISPLAYLIST_TYPE_TILE:
			XBUILDDISPLAYELEMENT_TILE DTile;
			memcpy(&DTile, l->Element[i].pData, sizeof(XBUILDDISPLAYELEMENT_TILE));
			float fX, fY;
			//Upload the texture if it's not upped
			nTile = DTile.dwTile;
			if(nTile == -1) continue;
			if(VOGL_nMesh[nTile] != -1)
			{
				//Ortho Deinit
				glMatrixMode(GL_PROJECTION);
				glPopMatrix();
				glMatrixMode(GL_MODELVIEW);
				glPopMatrix();
				glLoadIdentity();
				//Ortho Deinit

				glTranslatef(3.0f, -1.3f, -5.5f);
				glRotatef(-90.0f, 1.0f, 0.0f, 0.0f);
				glRotatef(-90.0f, 0.0f, 0.0f, 1.0f);
				glScalef(0.0003f, 0.0003f, 0.0003f);
				glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
				glBindTexture(GL_TEXTURE_2D, 0);
				glCallList(VOGL_nMesh[nTile]);

				//Ortho Reinit
				glMatrixMode(GL_PROJECTION);
				glPushMatrix();
				glLoadIdentity();
				if(l->dwReferentialX != 0xFFFFFFFF)
				{
					glOrtho(0, l->dwReferentialX, 0, l->dwReferentialY, -1, 1);
					nHeight = l->dwReferentialY;
				}
				else
				{
					glOrtho(0, VOGL_Width[VOGL_Config.nResolution], 0, VOGL_Height[VOGL_Config.nResolution], -1, 1);
					nHeight = VOGL_Height[VOGL_Config.nResolution];
				}
				glMatrixMode(GL_MODELVIEW);            
				glPushMatrix();
				glLoadIdentity();
				//Ortho Reinit
				continue;
			}
			if(!VOGL_TilesCreated[nTile])
			{
				VOGL_Tiles[nTile] = new XBUILD_ART_TILEDATA;
				XBUILD_ART_GetTileData(nTile, VOGL_Tiles[nTile]);
				UploadTexture(nTile);
				VOGL_TilesCreated[nTile] = 0xFF;
			}
			if(VOGL_Tiles[nTile]->Attributes & 0xC0)
			{
				nTile += XBUILD_ANIMATE_GetTileIncrement(VOGL_Tiles[nTile]);
			}
			if(!VOGL_TilesCreated[nTile])
			{
				VOGL_Tiles[nTile] = new XBUILD_ART_TILEDATA;
				XBUILD_ART_GetTileData(nTile, VOGL_Tiles[nTile]);
				UploadTexture(nTile);
				VOGL_TilesCreated[nTile] = 0xFF;
			}
			glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
			glBindTexture(GL_TEXTURE_2D, VOGL_nTexture[nTile]);
			glBegin(GL_QUADS);
			{
				glTexCoord2f(0.01f, 0.99f);	glVertex2i(DTile.dwX,								nHeight - (DTile.dwY));
				glTexCoord2f(0.99f, 0.99f);	glVertex2i(DTile.dwX + VOGL_Tiles[nTile]->Width,	nHeight - (DTile.dwY));
				glTexCoord2f(0.99f, 0.01f);	glVertex2i(DTile.dwX + VOGL_Tiles[nTile]->Width,	nHeight - (DTile.dwY + VOGL_Tiles[nTile]->Height));
				glTexCoord2f(0.01f, 0.01f);	glVertex2i(DTile.dwX,								nHeight - (DTile.dwY + VOGL_Tiles[nTile]->Height));
			}
			glEnd();
			break;
		case XBUILDDISPLAYLIST_TYPE_32BITS:
			unsigned int				nTexture;
			int							nWidth;
			int							nHeight;
			int							k;
			int							j;
			int							nPtr;
			unsigned _int32*			pNewData;
			unsigned _int32*			pData;
			XBUILDDISPLAYELEMENT_32BITS DStream;

			memcpy(&DStream, l->Element[i].pData, sizeof(XBUILDDISPLAYELEMENT_32BITS));
			nWidth = SearchForNextPower(l->dwReferentialX);
			nHeight = SearchForNextPower(l->dwReferentialY);
			pNewData = new unsigned _int32[nWidth * nHeight];
			pData = (unsigned _int32*)DStream.pStream;
			if(pData == NULL)
			{
				return;
			}
			//Ewww... way to slow
			gluScaleImage(GL_RGBA, l->dwReferentialX, l->dwReferentialY, GL_UNSIGNED_BYTE, pData, nWidth, nHeight, GL_UNSIGNED_BYTE, pNewData);
			//memset(pNewData, 0x00, nWidth * nHeight * 4);
			//memset(pNewData + ((nWidth * (nHeight - 4)), 0xFF, nWidth * 4 * 4);
			/*for(j = 0; j < l->dwReferentialY; j++)
			{
				for(k = 0; k < l->dwReferentialX; k++)
				{
					pNewData[nPtr + k] = pData[(j * l->dwReferentialX) + k] | 0xFF000000;
				}
				nPtr += nWidth;
			}*/
			glGenTextures(1, &nTexture);
			glBindTexture(GL_TEXTURE_2D, nTexture);
			glTexImage2D(GL_TEXTURE_2D, 0, 4, nWidth, nHeight, 0, GL_RGBA, GL_UNSIGNED_BYTE, pNewData);
			glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
			glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
			glBindTexture(GL_TEXTURE_2D, nTexture);
			glColor3f(1.0f, 1.0f, 1.0f);
			glBegin(GL_QUADS);
			{
				glTexCoord2f(0.0f, 0.0f);	glVertex2i(DStream.dwX,						200 - (DStream.dwY));
				glTexCoord2f(1.0f, 0.0f);	glVertex2i(DStream.dwX + l->dwReferentialX,	200 - (DStream.dwY));
				glTexCoord2f(1.0f, 1.0f);	glVertex2i(DStream.dwX + l->dwReferentialX,	200 - (DStream.dwY + l->dwReferentialY));
				glTexCoord2f(0.0f, 1.0f);	glVertex2i(DStream.dwX,						200 - (DStream.dwY + l->dwReferentialY));
			}
			glEnd();
			glDeleteTextures(1, &nTexture);
			delete pNewData;
			break;
		}
	}
	//Ortho Screen Deinit
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
	//Ortho Screen Deinit
	glDisable(GL_BLEND);
	glEnable(GL_DEPTH_TEST);
	delete VOGL_lpPalette;
	VOGL_lpPalette = NULL;
}


