#include "COMMON.H"

struct XBUILD_VOCSOUNDDATA
{
	unsigned _int16		SampleRate;
	unsigned _int8		CompressionType;
	unsigned _int32		PCMDATA_Size;
	unsigned _int8 *	PCMDATA;
};

#define XBUILD_SOUND_CHANNELS 64

LPDIRECTSOUND		SOUND_lpDS;
LPDIRECTSOUNDBUFFER	SOUND_lpDSB;
LPDIRECTSOUNDBUFFER	SOUND_lpDSPlay[XBUILD_SOUND_CHANNELS];
char				*SOUND_strFile[XBUILD_SOUND_CHANNELS];

WAVEFORMATEX SOUND_lpDSBwfx =
{
	WAVE_FORMAT_PCM,
	2,
	44100,
	176400,		//nChannels * Sample Rate
	2,
	16,
	sizeof(WAVEFORMATEX)
};

extern "C" int DLLEXPORT XBUILD_SOUND_Initialize(HWND hWnd)
{
	int i;
	DSBUFFERDESC dsbd;
	DirectSoundCreate(NULL, &SOUND_lpDS, NULL);
	if(SOUND_lpDS == NULL)
	{
		return 0;
	}
	SOUND_lpDS->SetCooperativeLevel(hWnd, DSSCL_PRIORITY | DSSCL_EXCLUSIVE);
	memset(&dsbd, 0, sizeof(DSBUFFERDESC));
	dsbd.dwSize = sizeof(DSBUFFERDESC);
	dsbd.dwFlags = DSBCAPS_PRIMARYBUFFER;
	SOUND_lpDS->CreateSoundBuffer(&dsbd, &SOUND_lpDSB, NULL);
	SOUND_lpDSB->Play(0, 0, DSBPLAY_LOOPING);
	SOUND_lpDSB->SetFormat(&SOUND_lpDSBwfx);
	for(i = 0; i < XBUILD_SOUND_CHANNELS; i++)
	{
		SOUND_lpDSPlay[i] = NULL;
		SOUND_strFile[i] = new char[64];
	}
	return 1;
}

extern "C" void DLLEXPORT XBUILD_SOUND_Destroy()
{
	int i;
	if(SOUND_lpDS != NULL)
	{
		if(SOUND_lpDSB != NULL)
		{
			SOUND_lpDSB->Release();
			SOUND_lpDSB = NULL;
		}
		for(i = 0; i < XBUILD_SOUND_CHANNELS; i++)
		{
			if(SOUND_lpDSPlay[i] != NULL)
			{
				SOUND_lpDSPlay[i]->Release();
				SOUND_lpDSPlay[i] = NULL;
			}
			delete SOUND_strFile[i];
		}
		SOUND_lpDS->Release();
		SOUND_lpDS = NULL;
	}
}

int SOUND_GetNextAvaliableChannel()
{
	int i;
	for(i = 0; i < XBUILD_SOUND_CHANNELS; i++)
	{
		if(SOUND_lpDSPlay[i] != NULL)
		{
			DWORD dwStatus;
			SOUND_lpDSPlay[i]->GetStatus(&dwStatus);
			if(!(dwStatus & DSBSTATUS_PLAYING))
			{
				SOUND_lpDSPlay[i]->Release();
				SOUND_lpDSPlay[i] = NULL;
				return i;
			}
		}
		else
		{
			return i;
		}
	}
	return -1;
}

void SOUND_FillWithPCMData(XBUILD_VOCSOUNDDATA *sound, unsigned _int8 *ptr, unsigned _int32 size)
{
	sound->SampleRate = 1000000 / (256 - ptr[0]);
	sound->CompressionType = ptr[1];
	if(sound->PCMDATA != NULL)
	{
		unsigned _int8 *pTemp;
		pTemp = new unsigned _int8[sound->PCMDATA_Size];
		memcpy(pTemp, sound->PCMDATA, sound->PCMDATA_Size);
		delete sound->PCMDATA;
		sound->PCMDATA = new unsigned _int8[sound->PCMDATA_Size + (size - 2)];
		memcpy(sound->PCMDATA, pTemp, sound->PCMDATA_Size);
		memcpy(sound->PCMDATA + (sound->PCMDATA_Size), ptr + 2, size - 2);
		sound->PCMDATA_Size += (size - 2);
		delete pTemp;
	}
	else
	{
		sound->PCMDATA = new unsigned _int8[size - 2];
		memcpy(sound->PCMDATA, ptr + 2, size - 2);
		sound->PCMDATA_Size = size - 2;
	}
}

void SOUND_FillWithPCMDataEx(XBUILD_VOCSOUNDDATA *sound, unsigned _int8 *ptr, unsigned _int32 size)
{
	sound->SampleRate = GETWORD(ptr, 0);
	ptr += 12;
	sound->PCMDATA = new unsigned _int8[size - 12];
	memcpy(sound->PCMDATA, ptr, size - 12);
	sound->PCMDATA_Size = size - 12;
}

void SOUND_FillVOCDataStruct(unsigned _int8 *lpBuffer, unsigned _int32 dwSize, XBUILD_VOCSOUNDDATA *d)
{
	unsigned _int8 command;
	unsigned _int32 datasize = 0;
	unsigned _int32 pos;
	memset(d, 0, sizeof(XBUILD_VOCSOUNDDATA));
	pos = GETWORD(lpBuffer, 0x14);
	while(1)
	{
		command = lpBuffer[pos];
		pos++;
		datasize = GETWORD(lpBuffer, pos) | (lpBuffer[pos + 2] << 16);
		pos += 3;
		switch(command)
		{
		case 0x00:
			//Terminator
			break;
		case 0x01:
			//Sound Data
			SOUND_FillWithPCMData(d, lpBuffer + pos, datasize);
			break;
		case 0x06:
			//Repeat
			break;
		case 0x09:
			//Sound Data (Extended Format)
			SOUND_FillWithPCMDataEx(d, lpBuffer + pos, datasize);
			break;
		}
		pos += datasize;
		if(pos >= dwSize)
		{
			break;
		}
	}
}

extern "C" void DLLEXPORT XBUILD_SOUND_SetVolume(int nSound, int nVolumePercent)
{
	float fVolume;
	long nVolume;
	fVolume = (float)nVolumePercent / 100.0f;
	fVolume = 1.0f - fVolume;
	fVolume *= DSBVOLUME_MIN;
	nVolume = (long)fVolume;
	nVolume += 3000;
//	nVolume /= 2;
	if(nSound != -1)
	{
		SOUND_lpDSPlay[nSound]->SetVolume(nVolume);
	}
}

extern "C" void DLLEXPORT XBUILD_SOUND_Stop(int nSound)
{
	if(nSound != -1)
	{
		SOUND_lpDSPlay[nSound]->Stop();
	}
}

extern "C" int DLLEXPORT XBUILD_SOUND_PlayEx(void *lpBuffer, DWORD dwSize, int nLoop)
{
	int nSpot;
	DSBUFFERDESC dsbd;
	WAVEFORMATEX wfx;
	void *soundaddr;
	DWORD size;
	XBUILD_VOCSOUNDDATA d;
	if(SOUND_lpDS == NULL)
	{
		return -1;
	}
	memset(&d, 0, sizeof(XBUILD_VOCSOUNDDATA));
	SOUND_FillVOCDataStruct((unsigned _int8*)lpBuffer, dwSize, &d);
	delete lpBuffer;
	//Fill in the DSBUFFERDESC thing
	memset(&dsbd, 0, sizeof(DSBUFFERDESC));
	dsbd.dwSize = sizeof(DSBUFFERDESC);
	dsbd.dwFlags = DSBCAPS_STATIC | DSBCAPS_CTRLVOLUME;
	dsbd.dwBufferBytes = d.PCMDATA_Size;
	//Fill in the WAVEFORMATEX thing
	memset(&wfx, 0, sizeof(WAVEFORMATEX));
	wfx.cbSize = sizeof(WAVEFORMATEX);
	wfx.wFormatTag = WAVE_FORMAT_PCM;
	wfx.nChannels = 1;
	wfx.nSamplesPerSec = d.SampleRate;
	wfx.wBitsPerSample = 8;
	wfx.nBlockAlign = 1;
	wfx.nAvgBytesPerSec = wfx.nSamplesPerSec;
	dsbd.lpwfxFormat = &wfx;
	//Create the buffer
	nSpot = SOUND_GetNextAvaliableChannel();
	SOUND_lpDS->CreateSoundBuffer(&dsbd, &SOUND_lpDSPlay[nSpot], NULL);
	SOUND_lpDSPlay[nSpot]->Lock(0, 0, (void**)&soundaddr, &size, NULL, NULL, DSBLOCK_ENTIREBUFFER);
	memmove(soundaddr, d.PCMDATA, size);
	SOUND_lpDSPlay[nSpot]->Unlock(soundaddr, size, NULL, 0);
	SOUND_lpDSPlay[nSpot]->SetCurrentPosition(0);
	if(nLoop)
	{
		SOUND_lpDSPlay[nSpot]->Play(0, 0, DSBPLAY_LOOPING);
	}
	else
	{
		SOUND_lpDSPlay[nSpot]->Play(0, 0, 0);
	}
	strcpy(SOUND_strFile[nSpot], "NONE");
	delete d.PCMDATA;
	return nSpot;
}

int SOUND_CheckCache(char *lpstrSound, int nLoop)
{
	int i;
	for(i = 0; i < XBUILD_SOUND_CHANNELS; i++)
	{
		if(SOUND_lpDSPlay[i] == NULL) continue;
		if(!strcmp(lpstrSound, SOUND_strFile[i]))
		{
			if(!nLoop)
			{
				SOUND_lpDSPlay[i]->SetCurrentPosition(0);
			}
			SOUND_lpDSPlay[i]->Play(0, 0, nLoop * DSBPLAY_LOOPING);
			return i;
		}
	}
	return -1;
}

extern "C" int DLLEXPORT XBUILD_SOUND_Play(char *lpstrSound, int nLoop)
{
	int nSpot;
	DSBUFFERDESC dsbd;
	WAVEFORMATEX wfx;
	void *soundaddr;
	DWORD size;
	unsigned _int8 *lpBuffer;
	unsigned _int32 dwSize = 0;
	XBUILD_VOCSOUNDDATA d;
	if(SOUND_lpDS == NULL)
	{
		return -1;
	}
	nSpot = SOUND_CheckCache(lpstrSound, nLoop);
	if(nSpot != -1)
	{
		return nSpot; 
	}
	memset(&d, 0, sizeof(XBUILD_VOCSOUNDDATA));
	XBUILD_GRPMANAGER_GetFileData(lpstrSound, &lpBuffer, &dwSize);
	SOUND_FillVOCDataStruct(lpBuffer, dwSize, &d);
	delete lpBuffer;
	//Fill in the DSBUFFERDESC thing
	memset(&dsbd, 0, sizeof(DSBUFFERDESC));
	dsbd.dwSize = sizeof(DSBUFFERDESC);
	dsbd.dwFlags = DSBCAPS_STATIC | DSBCAPS_CTRLVOLUME;
	dsbd.dwBufferBytes = d.PCMDATA_Size;
	//Fill in the WAVEFORMATEX thing
	memset(&wfx, 0, sizeof(WAVEFORMATEX));
	wfx.cbSize = sizeof(WAVEFORMATEX);
	wfx.wFormatTag = WAVE_FORMAT_PCM;
	wfx.nChannels = 1;
	wfx.nSamplesPerSec = d.SampleRate;
	wfx.wBitsPerSample = 8;
	wfx.nBlockAlign = 1;
	wfx.nAvgBytesPerSec = wfx.nSamplesPerSec;
	dsbd.lpwfxFormat = &wfx;
	//Create the buffer
	nSpot = SOUND_GetNextAvaliableChannel();
	try
	{
		SOUND_lpDS->CreateSoundBuffer(&dsbd, &SOUND_lpDSPlay[nSpot], NULL);
		SOUND_lpDSPlay[nSpot]->Lock(0, 0, &soundaddr, &size, NULL, NULL, DSBLOCK_ENTIREBUFFER);
		memcpy(soundaddr, d.PCMDATA, size);
		SOUND_lpDSPlay[nSpot]->Unlock(soundaddr, size, NULL, 0);
		SOUND_lpDSPlay[nSpot]->SetCurrentPosition(0);
		if(nLoop)
		{
			SOUND_lpDSPlay[nSpot]->Play(0, 0, DSBPLAY_LOOPING);
		}
		else
		{
			SOUND_lpDSPlay[nSpot]->Play(0, 0, 0);
		}
	}
	catch(...)
	{
		return -1;
	}
	strcpy(SOUND_strFile[nSpot], lpstrSound);
	delete d.PCMDATA;
	return nSpot;
}