#ifndef _XBUILD_H_
#define _XBUILD_H_

#include "XBUILDREN.H"

#define MAXSECTORS 1024
#define MAXWALLS 8192
#define MAXSPRITES 4096

#define MAXTILES 9216
#define MAXSTATUS 1024
#define MAXPLAYERS 16
#define MAXXDIM 1600
#define MAXYDIM 1200
#define MAXPALOOKUPS 256
#define MAXPSKYTILES 256
#define MAXSPRITESONSCREEN 1024

#define CLIPMASK0 (((1L)<<16)+1L)
#define CLIPMASK1 (((256L)<<16)+64L)

#define MAXSTATE	0xBE
#define MAXACTOR	0xF0

struct BLK_DEFINE
{
	char strSymbol[64];
	int nValue;
};

struct BLK_DEFINEQUOTE
{
	int nNumber;
	char strQuote[64];
};

struct BLK_DEFINESOUND
{
	int		nNumber;
	char	strFilename[32];
	int		nPitchVariation1;
	int		nPitchVariation2;
	int		nPriority;
	int		nStatus;
	int		nVolumeAdjust;
};

struct BLK_DEFINEVOLUMENAME
{
	int		nNumber;
	char	strName[32];
};

struct BLK_DEFINESKILLNAME
{
	int		nNumber;
	char	strName[32];
};

struct BLK_DEFINELEVELNAME
{
	int		nEpisode;
	int		nLevel;
	char	strFilename[32];
	char	strPar[10];
	char	str3DRealmsPar[10];
	char	strLevelname[32];
};

struct BLK_MUSIC
{
	int		nEpisode;
	char	strFilename01[32];
	char	strFilename02[32];
	char	strFilename03[32];
	char	strFilename04[32];
	char	strFilename05[32];
	char	strFilename06[32];
	char	strFilename07[32];
	char	strFilename08[32];
	char	strFilename09[32];
	char	strFilename10[32];
	char	strFilename11[32];
};

struct BLK_GAMESTARTUP
{
	int		nVisibility;
	int		nImpactDamage;
	int		nMaxPlayerHealth;
	int		nArmorHealth;
	int		nActorRespawnTime;
	int		nItemRespawnTime;
	int		nRunningSpeed;
	int		nGravitationnalConstant;
	int		nRPGBlastRadius;
	int		nPipebombBlastRadius;
	int		nShrinkerBlastRadius;
	int		nTripbombBlastRadius;
	int		nMorterBlastRadius;
	int		nBounceMineBlastRadius;
	int		nSeenineBlastRadius;
	int		nMaxPistolAmmo;
	int		nMaxShotgunAmmo;
	int		nMaxChaingunAmmo;
	int		nMaxRPGAmmo;
	int		nMaxPipebombAmmo;
	int		nMaxShrinkerAmmo;
	int		nMaxDevastatorAmmo;
	int		nMaxTripbombAmmo;
	int		nMaxFreezeAmmo;
	int		nMaxGrowAmmo;
	int		nCameraDestructible;
	int		nNumberFreezeBounces;
	int		nFreezerHurtsOwner;
	int		nQSize;
	int		nTripbombLaserMode;
};

struct BLK_ACTOR
{
	int		nSprite;
	int		nHealth;
	char	strCode[0x3000];
};

struct BLK_STATE
{
	char	strName[256];
	char	strCode[0x3000];
};

struct XBUILD_CONSCRIPT_CONTEXT
{
	BLK_DEFINE				Define[0x600];
	BLK_DEFINEQUOTE			DefineQuote[0x80];
	BLK_DEFINESOUND			DefineSound[400];
	BLK_DEFINEVOLUMENAME	DefineVolumename[10];
	BLK_DEFINELEVELNAME		DefineLevelname[50];
	BLK_DEFINESKILLNAME		DefineSkillname[10];
	BLK_MUSIC				Music[5];
	BLK_GAMESTARTUP			Gamestartup;
	BLK_ACTOR				Actor[MAXACTOR];
	BLK_STATE				State[MAXSTATE];
	int						nDefineCount;
	int						nDefineQuoteCount;
	int						nDefineSoundCount;
	int						nDefineVolumenameCount;
	int						nDefineLevelnameCount;
	int						nDefineSkillnameCount;
	int						nMusicCount;
	int						nActorCount;
	int						nStateCount;
	int						nGamestartupCount;
};

//MAP Loader Types
struct XBUILD_MAP_MAPINFO
{
	unsigned	_int32	Version;
	signed		_int32	StartX;
	signed		_int32	StartY;
	signed		_int32	StartZ;
	unsigned	_int16	StartA;
	unsigned	_int16	FirstSector;
	unsigned	_int16	SectorNumber;
	unsigned	_int16	WallNumber;
	unsigned	_int16	SpriteNumber;
};

struct XBUILD_MAP_SECTOR
{
	unsigned	_int16	WallPointer;
	unsigned	_int16	WallNumber;
	signed		_int32	CZ;					//Ceiling Z position
	signed		_int32	FZ;					//Floor Z position
	unsigned	_int16	CStat;				//Ceiling Status
	unsigned	_int16	FStat;				//Floor Status
	unsigned	_int16	CPic;				//Ceiling Picture Number
	unsigned	_int16	CHei;				//Ceiling Slope
	signed		_int8	CShade;				//Ceiling Shade
	unsigned	_int8	CPal;				//Ceiling Palette
	unsigned	_int8	CPanX;				//Ceiling Texture Pan X
	unsigned	_int8	CPanY;				//Ceiling Texture Pan Y
	unsigned	_int16	FPic;				//Floor Picture Number
	unsigned	_int16	FHei;				//Floor Slope
	signed		_int8	FShade;				//Floor Shade
	unsigned	_int8	FPal;
	unsigned	_int8	FPanX;
	unsigned	_int8	FPanY;
	unsigned	_int8	Visibility;
	unsigned	_int8	Filler;
	unsigned	_int16	LoTag;
	unsigned	_int16	HiTag;
	unsigned	_int16	Extra;
};

struct XBUILD_MAP_WALL
{
	signed		_int32	X;
	signed		_int32	Y;
	unsigned	_int16	Point2;
	unsigned	_int16	NextWall;
	signed		_int16	NextSector;
	unsigned	_int16	CStat;
	unsigned	_int16	Pic;
	unsigned	_int16	OverPic;
	signed		_int8	Shade;
	unsigned	_int8	Pal;
	unsigned	_int8	RepX;
	unsigned	_int8	RepY;
	unsigned	_int8	PanX;
	unsigned	_int8	PanY;
	unsigned	_int16	LoTag;
	unsigned	_int16	HiTag;
	unsigned	_int16	Extra;
};

struct XBUILD_MAP_SPRITE
{
	signed		_int32	X;
	signed		_int32	Y;
	signed		_int32	Z;
	unsigned	_int16	CStat;
	unsigned	_int16	Pic;
	signed		_int8	Shade;
	unsigned	_int8	Palette;
	unsigned	_int8	Clipdist;
	unsigned	_int8	Filler;
	unsigned	_int8	RepX;
	unsigned	_int8	RepY;
	signed		_int8	OffX;
	signed		_int8	OffY;
	unsigned	_int16	SectorNum;
	unsigned	_int16	StatNum;
	unsigned	_int16	Angle;
	unsigned	_int16	Owner;
	unsigned	_int16	VelX;
	unsigned	_int16	VelY;
	unsigned	_int16	VelZ;
	unsigned	_int16	LoTag;
	unsigned	_int16	HiTag;
	unsigned	_int16	Extra;
};

struct XBUILD_MAP
{
	XBUILD_MAP_MAPINFO	Mapinfo;
	XBUILD_MAP_SECTOR	Sector[MAXSECTORS];
	XBUILD_MAP_WALL		Wall[MAXWALLS];
	XBUILD_MAP_SPRITE	Sprite[MAXSPRITES];
};

//ART Loader Types
struct XBUILD_ART_TILEDATA
{
	unsigned _int16 Width;
	unsigned _int16 Height;
	unsigned _int32 Attributes;
	unsigned _int8  *lpData;
};

struct XBUILD_INCREMENT
{
	int *pValue;
	int nUnitsPerSlice;
	int nMaxValue;
	int nMsPerSlice;
	int nReserved;
};

struct XBUILD_CONSCRIPTACTOR_CONTEXT
{
	XBUILD_MAP_SPRITE*	pSprite;
	unsigned short		wPC;
	unsigned short		wTrue;
	unsigned short		wMove;
	unsigned short		wHP;
	unsigned short		wPlayerStatus;
	unsigned short		wCount;
	unsigned short		wSpawnedBy;
	unsigned short		wConditions;
	unsigned char		bHitByWeapon;
	unsigned char		pCode[0x100];
};

extern "C" int XBUILD_Initialize(char*, char*);

//ART Loader/Converter Functions
extern "C" void		XBUILD_ART_Destroy();
extern "C" int		XBUILD_ART_Initialize();
extern "C" void		XBUILD_ART_GetTileData(int, XBUILD_ART_TILEDATA*);
extern "C" void		XBUILD_ART_GetTileDimensions(int, unsigned short*, unsigned short*);
extern "C" void		XBUILD_ART_GetPalette(void*);

//Console Control Functions
extern "C" void		XBUILD_CONSOLE_PrintFormatted(const char*, ...);
extern "C" void		XBUILD_CONSOLE_Clear();
extern "C" void		XBUILD_CONSOLE_ShowTextConsole(int);
extern "C" void		XBUILD_CONSOLE_HideTextConsole();
extern "C" void		XBUILD_CONSOLE_GetGraphicConsole(XBUILDDISPLAYLIST*);
extern "C" void		XBUILD_CONSOLE_SetGraphicConsoleParams(int, int);
extern "C" void		XBUILD_CONSOLE_SetConsoleCallback(void*, void*);

//GRP File Manager Functions
extern "C" int		XBUILD_GRPMANAGER_GetFileData(char*, unsigned _int8**, unsigned _int32*);
extern "C" int		XBUILD_GRPMANAGER_ExtractFile(char*, char*);
extern "C" void		XBUILD_GRPMANAGER_SetSecondSearchDir(char*);

//CON Compiler Functions
#define XBUILD_COMPILER_MODE_V13		1
#define XBUILD_COMPILER_MODE_V1415		2

extern "C" int		XBUILD_CONSCRIPT_Compile(const char*, XBUILD_CONSCRIPT_CONTEXT*);
extern "C" void		XBUILD_CONSCRIPT_SetCompilerMode(int);

//CON Actor Compiler Functions
extern "C" void		XBUILD_CONACTORCOMPILER_SetConscriptContext(XBUILD_CONSCRIPT_CONTEXT*);
extern "C" void		XBUILD_CONACTORCOMPILER_Assemble(int, unsigned char*);

//CON Virtual Machine Functions
extern "C" void		XBUILD_CONVM_Disassemble(char*, unsigned char*, unsigned short, unsigned short);
extern "C" void		XBUILD_CONVM_Execute(XBUILD_CONSCRIPTACTOR_CONTEXT *pC);
extern "C" void		XBUILD_CONVM_SetFunctionCallback(void*);

//CON Virtual Machine Commands
#define VM_ADDHEALTH	0xB000
#define VM_GETHEALTH	0xB001
#define VM_GLOBALSOUND	0xB002
#define VM_QUOTE		0xB003
#define VM_GETDISTANCE	0xB004
#define VM_CANSEE		0xB005
#define VM_ADDAMMO		0xB006

//GFX Renderer/Manager Functions
#define				XBUILD_GFX_LAYER_3DMAP		0x01
#define				XBUILD_GFX_LAYER_CONSOLE	0x02
#define				XBUILD_GFX_LAYER_STATUS		0x04

extern "C" int		XBUILD_GFX_AddTileToDisplayList(int, int, int, XBUILDDISPLAYLIST*, int);
extern "C" void		XBUILD_GFX_Destroy();
extern "C" void*	XBUILD_GFX_GetMapAddress();
extern "C" HWND		XBUILD_GFX_Initialize(char*);
extern "C" void		XBUILD_GFX_SetMap(XBUILD_MAP*);
extern "C" int		XBUILD_GFX_Update(int, int, int, int, int);
extern "C" void		XBUILD_GFX_SetStatus(int);
extern "C" void		XBUILD_GFX_SetDisplayList(int, XBUILDDISPLAYLIST*);
extern "C" void		XBUILD_GFX_SetMeshForTile(int, char*, float);
extern "C" void		XBUILD_GFX_SetSystemCommandCallback(void*, void*);
extern "C" void		XBUILD_GFX_TakeScreenShot(char*);
extern "C" void		XBUILD_GFX_SaveImageAsTGA(char*, int, int, int, unsigned _int8*);

//ANIMATE Functions
extern "C" void		XBUILD_ANIMATE_Initialize();
extern "C" void		XBUILD_ANIMATE_UpdateSector(long, long, int*);
extern "C" void		XBUILD_ANIMATE_SetCurrentPosition(int, int, int);
extern "C" BOOL		XBUILD_ANIMATE_IsPositionValid(int, int, int, int);
extern "C" int		XBUILD_ANIMATE_HasCurrentSectorChanged();
extern "C" void		XBUILD_ANIMATE_ChangeCurrentSector(int);
extern "C" int		XBUILD_ANIMATE_GetFloorHeight(short, long, long);
extern "C" int		XBUILD_ANIMATE_GetCeilingHeight(short, long, long);
extern "C" int		XBUILD_ANIMATE_NearTag(long, long, long, short, short, short*, short*, short*, long*, long, char);
extern "C" int		XBUILD_ANIMATE_GetCurrentSector();
extern "C" int		XBUILD_ANIMATE_GetTileIncrement(XBUILD_ART_TILEDATA*);
extern "C" int		XBUILD_ANIMATE_LookForSpriteInSector(int, int, int);
extern "C" int		XBUILD_ANIMATE_LookForSpriteWithSameTag(int, int, int);
extern "C" float	XBUILD_ANIMATE_BuildToDegree(int);
extern "C" float	XBUILD_ANIMATE_BuildToRadians(int);
extern "C" int		XBUILD_ANIMATE_ClipMove(long*, long*, long*, short*, long, long, long, long, long, unsigned long);
extern "C" int		XBUILD_ANIMATE_GetValueFromSinTable(int);
extern "C" int		XBUILD_ANIMATE_CanSee(long, long, long, short, long, long, long, short);
extern "C" void		XBUILD_ANIMATE_RotatePoint(long, long, long, long, short, long*, long*);
extern "C" int		XBUILD_ANIMATE_NextSectorNeighborZ(short, long, short, short);
extern "C" int		XBUILD_ANIMATE_HitScan(long, long, long, short, long, long, long, short*, short*, short*, long*, long*, long*, unsigned long);

//MAP Loader Function
extern "C" int		XBUILD_MAP_Load(char*, XBUILD_MAP*);

//MUSIC Player Functions
#define				XBUILD_MUSIC_DRIVER_MCI		0x00
#define				XBUILD_MUSIC_DRIVER_DMUS	0x01
#define				XBUILD_MUSIC_DRIVER_CD		0x02

extern "C" void		XBUILD_MUSIC_Play(char*, int);
extern "C" void		XBUILD_MUSIC_Stop();

//SOUND Player Functions
extern "C" int		XBUILD_SOUND_Initialize(HWND);
extern "C" void		XBUILD_SOUND_Destroy();
extern "C" int		XBUILD_SOUND_Play(char*, int);
extern "C" void		XBUILD_SOUND_Stop(int);
extern "C" void		XBUILD_SOUND_SetVolume(int, int);

//INCREMENTATOR Functions
extern "C" int		XBUILD_INCREMENTATOR_Add(int*, int, int, void*, void*);
extern "C" void		XBUILD_INCREMENTATOR_Destroy();
extern "C" void		XBUILD_INCREMENTATOR_Initialize();
extern "C" int		XBUILD_INCREMENTATOR_IsIncrementDead(int);
extern "C" void		XBUILD_INCREMENTATOR_SetMaxValue(int, int);
extern "C" void		XBUILD_INCREMENTATOR_Update();

//RTSMANAGER Functions
extern "C" void		XBUILD_RTSMANAGER_Play(int);

#endif